#!/usr/bin/env pike

void test_callback(mixed ... args) {
  write("args==%O\n",args);
}

array actions=({
 ([ "name": "FileMenu", "label": "_File" ]),
 ([ "name": "PreferencesMenu", "label": "_Preferences" ]),
 ([ "name": "ColorMenu", "label": "_Color" ]),
 ([ "name": "ShapeMenu", "label": "_Shape" ]),
 ([ "name": "HelpMenu", "label": "_Help" ]),
 ([ "name": "New", "stock_id": GTK2.STOCK_NEW, "label": "_New",
    "accelerator": "<control>N", "tooltip": "Create a new file",
    "callback": test_callback, "data": "NEW" ]),
 ([ "name": "Open", "stock_id": GTK2.STOCK_OPEN, "label": "_Open",
    "accelerator": "<control>O", "tooltip": "Open a file",
    "callback": test_callback, "data": "OPEN" ]),
 ([ "name": "Save", "stock_id": GTK2.STOCK_SAVE, "label": "_Save",
    "accelerator": "<control>S", "tooltip": "Save current file",
    "callback": test_callback, "data": "SAVE" ]),
 ([ "name": "SaveAs", "stock_id": GTK2.STOCK_SAVE, "label": "_Save _As...",
    "tooltip": "Save to a file", "callback": test_callback,
    "data": "SAVEAS" ]),
 ([ "name": "Quit", "stock_id": GTK2.STOCK_QUIT, "label": "_Quit",
    "accelerator": "<control>Q", "tooltip": "Quit",
    "callback": lambda() {exit(0); }, "data": "QUIT" ]),
 ([ "name": "About", "label": "_About", "accelerator": "<control>A",
    "callback": test_callback, "data": "ABOUT" ])
});

array toggle_actions=({
 ([ "name": "Bold", "stock_id": GTK2.STOCK_BOLD, "label": "_Bold",
    "accelerator": "<control>B", "tooltip": "Bold",
    "callback": test_callback, "data": "BOLD" ])
});

array color_radio_actions=({
 ([ "name": "Red", "label": "_Red", "accelerator": "<control>R",
    "tooltip": "Blood", "value": 0 ]),
 ([ "name": "Green", "label": "_Green", "accelerator": "<control>G",
    "tooltip": "Grass", "value": 1 ]),
 ([ "name": "Blue", "label": "_Blue", "accelerator": "<control>B",
    "tooltip": "Sky", "value": 2 ])
});

array shape_radio_actions=({
 ([ "name": "Square", "label": "_Square", "accelerator": "<control>S",
    "tooltip": "Square", "value": 0 ]),
 ([ "name": "Rectangle", "label": "_Rectangle", "accelerator": "<control>R",
    "tooltip": "Rectangle", "value": 1 ]),
 ([ "name": "Oval", "label": "_Oval", "accelerator": "<control>O",
    "tooltip": "Egg", "value": 2 ])
});

string ui_info=
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='New'/>"
"      <menuitem action='Open'/>"
"      <menuitem action='Save'/>"
"      <menuitem action='SaveAs'/>"
"      <separator/>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='PreferencesMenu'>"
"      <menu action='ColorMenu'>"
"        <menuitem action='Red'/>"
"        <menuitem action='Green'/>"
"        <menuitem action='Blue'/>"
"      </menu>"
"      <menu action='ShapeMenu'>"
"        <menuitem action='Square'/>"
"        <menuitem action='Rectangle'/>"
"        <menuitem action='Oval'/>"
"      </menu>"
"      <menuitem action='Bold'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"  <toolbar name='ToolBar'>"
"    <toolitem action='Open'/>"
"    <toolitem action='Quit'/>"
/*
"    <separator action='Sep1'/>"
"    <toolitem action='Logo'/>"
*/
"  </toolbar>"
"</ui>";

#define GAUGE_THIS(X) t=gauge { (X); }; write("time for "+#X+": %f\n",t)

int main(int argc, array argv) {
  float t;
  object w,gt1,ui,vb;
  int ui_info_id;
  GAUGE_THIS(GTK2.setup_gtk());
  GAUGE_THIS(w=GTK2.Window(GTK2.WINDOW_TOPLEVEL));
  GAUGE_THIS(gt1=GTK2.ActionGroup("normal"));
  GAUGE_THIS(gt1->add_actions(actions));
  GAUGE_THIS(gt1->add_toggle_actions(toggle_actions));
  GAUGE_THIS(gt1->add_radio_actions(color_radio_actions,test_callback,"COLOR"));
  GAUGE_THIS(gt1->add_radio_actions(shape_radio_actions,test_callback,"SHAPE"));
  GAUGE_THIS(ui=GTK2.UiManager());
  GAUGE_THIS(ui->insert_action_group(gt1,0));
  GAUGE_THIS(w->add_accel_group(ui->get_accel_group()));
  GAUGE_THIS(w->set_title("UI Manager"));
  GAUGE_THIS(ui_info_id=ui->add_ui_from_string(ui_info));
  GAUGE_THIS(w->set_border_width(0));
  GAUGE_THIS(vb=GTK2.Vbox(0,0));
  GAUGE_THIS(vb->add(ui->get_widget("/MenuBar")));
  GAUGE_THIS(vb->add(ui->get_widget("/ToolBar")));
  GAUGE_THIS(w->add(vb));
  GAUGE_THIS(w->show_all());
  GAUGE_THIS(w->signal_connect("destroy",lambda(){ exit(0); },0));
  return -1;
}

